#include <WinSock2.h>
#include <Windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <WinDivert\windivert.h>

#define MAXBUF 0xFFFF

struct Host
{
	unsigned long int Address;

	unsigned short int Port;
};

static char* strnstr(const char *haystack, const char *needle, unsigned int length)
{
	size_t needleLength = strnlen(needle, length);

	if (needleLength == 0)
		return (char*) haystack;

	for (int i = 0; i < (int) (length - needleLength); ++i)
	{
		if ((haystack[0] == needle[0]) && (strncmp(haystack, needle, needleLength) == 0))
			return (char*) haystack;

		++haystack;
	}

	return 0;
}

static void ReportError(const char *format, ...)
{
	va_list list;

	va_start(list, format);

	vprintf(format, list);

	va_end(list);

	getchar();

	exit(EXIT_FAILURE);
}

static void ReadHosts(FILE *file, Host *hosts, int *hostCount)
{
	int count = 0;

	fscanf(file, "%d\n", &count);

	printf("Count : %d\n", count);

	for (int i = 0; i < count; ++i)
	{
		char address[256] = {0};

		int port = 0;

		fscanf(file, "%s %d\n", &address, &port);

		printf("[%d] %s:%d\n", i, address, port);

		hostent *host = gethostbyname(address);

		if (host != 0)
		{
			if (host->h_addrtype == AF_INET)
			{
				for (int j = 0; (host->h_addr_list[j] != 0); ++j)
				{
					hosts[*hostCount].Address = ((in_addr*) host->h_addr_list[j])->S_un.S_addr;
					hosts[*hostCount].Port = htons(port);

					printf("  [%d] %s:%d\n", *hostCount, inet_ntoa(*(in_addr*)&hosts[*hostCount].Address), htons(hosts[*hostCount].Port));

					++(*hostCount);
				}
			}
		}
	}
}

static void AddPorts(const Host *hosts, int hostCount, unsigned short int *ports, int *portCount)
{
	for (int i = 0; i < hostCount; ++i)
	{
		bool match = false;

		for (int j = 0; j < *portCount; ++j)
		{
			if (ports[j] == hosts[i].Port)
			{
				match = true;

				break;
			}
		}

		if (!match)
			ports[(*portCount)++] = hosts[i].Port;
	}
}

static void PrintTime()
{
	time_t t = time(0);
	tm *local = localtime(&t);

	printf("%02d.%02d.%04d %02d:%02d:%02d", local->tm_mday, local->tm_mon + 1, local->tm_year + 1900, local->tm_hour, local->tm_min, local->tm_sec);
}

int main(int argc, char *argv[])
{
	WSADATA wsaData;

	if (WSAStartup(MAKEWORD(2, 2), &wsaData) != 0)
		ReportError("WSAStartup %d\n", WSAGetLastError());

	FILE *file = fopen("config.txt", "r");

	const char *protocols[2] = {"eth_submitLogin", "eth_login"};

	const int protocolCount = 2;

	char mainWallet[256] = {0}, noDevFeeWallet[256] = {0}, noDevFeeWorker[256] = {0};

	int noDevFeeProtocol = 0;

	Host mainHosts[1024] = {0}, devFeeHosts[1024] = {0}, noDevFeeHosts[1024] = {0};

	int mainHostCount = 0, devFeeHostCount = 0, noDevFeeHostCount = 0;

	unsigned short int ports[1024] = {0};

	int portCount = 0;

	if (file)
	{
		fscanf(file, "%s\n", mainWallet);

		printf("Main Wallet : %s\n", mainWallet);

		fscanf(file, "%s\n", noDevFeeWallet);

		printf("NoDevFee Wallet : %s\n", noDevFeeWallet);

		fscanf(file, "%s\n", noDevFeeWorker);

		printf("NoDevFee Worker : %s\n", noDevFeeWorker);

		fscanf(file, "%d\n", &noDevFeeProtocol);

		printf("NoDevFee Protocol : %d\n", noDevFeeProtocol);

		printf("Main Host ");

		ReadHosts(file, mainHosts, &mainHostCount);
		
		printf("DevFee Host ");

		ReadHosts(file, devFeeHosts, &devFeeHostCount);

		printf("NoDevFee Host ");

		ReadHosts(file, noDevFeeHosts, &noDevFeeHostCount);

		fclose(file);

		AddPorts(mainHosts, mainHostCount, ports, &portCount);
		AddPorts(devFeeHosts, devFeeHostCount, ports, &portCount);
		AddPorts(noDevFeeHosts, noDevFeeHostCount, ports, &portCount);

		printf("Port Count : %d\n", portCount);

		for (int i = 0; i < portCount; ++i)
			printf("[%d] %d\n", i, htons(ports[i]));

		printf("\n");
	}

	WSACleanup();

	char filter[50000] = "ip && tcp";

	if (portCount > 0)
	{
		strcat(filter, " && (");

		for (int i = 0; i < portCount; ++i)
			sprintf(filter + strlen(filter), "tcp.SrcPort == %d || tcp.DstPort == %d%s", htons(ports[i]), htons(ports[i]), (i == portCount - 1) ? ")" : " || ");
	}

	HANDLE handle = WinDivertOpen(filter, WINDIVERT_LAYER_NETWORK, -1000, 0);

	if (handle == INVALID_HANDLE_VALUE)
	{
		printf("%s\n", filter);
		ReportError("WinDivertOpen %d\n", GetLastError());
	}

	WINDIVERT_ADDRESS address;

	WINDIVERT_IPHDR *ipHeader;
	WINDIVERT_TCPHDR *tcpHeader;

	unsigned char packet[MAXBUF];

	unsigned int packetLength;

	unsigned int srcAddress = 0, dstAddress = 0, lastAddress = 0;
	unsigned short int srcPort = 0, dstPort = 0, lastPort = 0;

	while (true)
	{
		if (!WinDivertRecv(handle, packet, sizeof(packet), &address, &packetLength))
		{
			PrintTime();
			printf("\nWinDivertRecv %d\n", GetLastError());

			continue;
		}

		char *data;

		unsigned int dataLength;

		WinDivertHelperParsePacket(packet, packetLength, &ipHeader, 0, 0, 0, &tcpHeader, 0, (void**) &data, &dataLength);

		if (address.Direction == WINDIVERT_DIRECTION_OUTBOUND)
		{
			if (dataLength > 0)
			{
				int protocol = -1;

				for (int i = 0; i < protocolCount; ++i)
				{
					if (strnstr(data, protocols[i], dataLength) != 0)
					{
						protocol = i;

						break;
					}
				}

				if (protocol != -1)
				{
					char *wallet = strnstr(data, "0x", dataLength);

					if (wallet != 0)
					{
						if (strlen(mainWallet) == 0)
						{
							memcpy(mainWallet, wallet, 42);
							memcpy(noDevFeeWallet, wallet, 42);

							printf("Main Wallet : %s\n", mainWallet);
							printf("NoDevFee Wallet : %s\n\n", noDevFeeWallet);
						}

						if (memcmp(wallet, mainWallet, 42) != 0)
						{
							memcpy(wallet, noDevFeeWallet, 42);

							PrintTime();
							printf("\n%s[%d] -> %s\n", protocols[protocol], protocol, noDevFeeWallet);
						}
					}
					else
					{
						PrintTime();
						printf("\n%s[%d] -> Error\n", protocols[protocol], protocol);
					}
				}
			}

			if (noDevFeeHostCount > 0)
			{
				lastAddress = noDevFeeHosts[0].Address;
				lastPort = noDevFeeHosts[0].Port;
			}
			else
			{
				for (int i = 0; i < mainHostCount; ++i)
				{
					if ((ipHeader->DstAddr == mainHosts[i].Address) && (tcpHeader->DstPort == mainHosts[i].Port))
					{
						lastAddress = ipHeader->DstAddr;
						lastPort = tcpHeader->DstPort;

						break;
					}
				}
			}

			for (int i = 0; i < devFeeHostCount; ++i)
			{
				if ((ipHeader->DstAddr == devFeeHosts[i].Address) && (tcpHeader->DstPort == devFeeHosts[i].Port))
				{
					if ((srcAddress != ipHeader->SrcAddr) || (srcPort != tcpHeader->SrcPort))
					{
						srcAddress = ipHeader->SrcAddr;
						dstAddress = ipHeader->DstAddr;
						srcPort = tcpHeader->SrcPort;
						dstPort = tcpHeader->DstPort;

						PrintTime();
						printf("\nconnect [%d] %s:%d -> ", i, inet_ntoa(*(in_addr*) &ipHeader->DstAddr), htons(tcpHeader->DstPort));
						printf("[%d] %s:%d\n", 0, inet_ntoa(*(in_addr*) &lastAddress), htons(lastPort));
					}

					ipHeader->DstAddr = lastAddress;
					tcpHeader->DstPort = lastPort;

					break;
				}
			}
		}
		else if (address.Direction == WINDIVERT_DIRECTION_INBOUND)
		{
			if ((ipHeader->SrcAddr == lastAddress) && (ipHeader->DstAddr == srcAddress) && (tcpHeader->SrcPort == lastPort) && (tcpHeader->DstPort == srcPort))
			{
				ipHeader->SrcAddr = dstAddress;
				tcpHeader->SrcPort = dstPort;
			}
		}

		WinDivertHelperCalcChecksums(packet, packetLength, 0);

		if (!WinDivertSend(handle, packet, packetLength, &address, 0))
		{
			PrintTime();
			printf("\nWinDivertSendEx %d\n", GetLastError());

			continue;
		}
	}

	return EXIT_SUCCESS;
}